@@ -21,40 +21,24 @@
 
 #include <config.h>
 #include <sys/types.h>
+
+/* XXX: some of these includes probably don't belong in a common header file */
 #include <sys/stat.h>
 #include <stdio.h>
-
-#if defined(HAVE_STRING_H) || defined(STDC_HEADERS)
 #include <string.h>
-#else
-#include <strings.h>
-#ifndef strchr
-#define strchr index
-#endif
-#ifndef strrchr
-#define strrchr rindex
-#endif
-#endif
-
 #include <errno.h>
+#include <stdlib.h>
+#include <unistd.h>
+#include <time.h>
+#include <limits.h>
+#include <stdbool.h>
+
+
+
 #ifndef errno
 extern int errno;
 #endif
 
-#ifdef STDC_HEADERS
-#include <stdlib.h>
-#endif
-
-/* The presence of unistd.h is assumed by gnulib these days, so we 
- * might as well assume it too. 
- */
-#include <unistd.h>
-
-#include <time.h>
-
-#if HAVE_LIMITS_H
-# include <limits.h>
-#endif
 #ifndef CHAR_BIT
 # define CHAR_BIT 8
 #endif
@@ -62,27 +46,21 @@ extern int errno;
 #if HAVE_INTTYPES_H
 # include <inttypes.h>
 #endif
+typedef bool boolean;
 
 #include "regex.h"
 #include "timespec.h"
 #include "buildcmd.h"
+#include "quotearg.h"
+#include "stat_.h" /* S_ISUID etc. */
 
+/* These days we will assume ANSI/ISO C protootypes work on our compiler. */
+#define PARAMS(Args) Args
 
-#ifndef S_IFLNK
-#define lstat stat
+#ifndef ATTRIBUTE_NORETURN
+# define ATTRIBUTE_NORETURN __attribute__ ((__noreturn__))
 #endif
 
-# ifndef PARAMS
-#  if defined PROTOTYPES || (defined __STDC__ && __STDC__)
-#   define PARAMS(Args) Args
-#  else
-#   define PARAMS(Args) ()
-#  endif
-# endif
-
-int lstat PARAMS((const char *__path, struct stat *__statbuf));
-int stat PARAMS((const char *__path, struct stat *__statbuf));
-
 int optionl_stat PARAMS((const char *name, struct stat *p));
 int optionp_stat PARAMS((const char *name, struct stat *p));
 int optionh_stat PARAMS((const char *name, struct stat *p));
@@ -91,48 +69,6 @@ int debug_stat   PARAMS((const char *file, struct stat *bufp));
 void set_stat_placeholders PARAMS((struct stat *p));
 int get_statinfo PARAMS((const char *pathname, const char *name, struct stat *p));
 
-#if ! defined HAVE_FCHDIR && ! defined fchdir
-# define fchdir(fd) (-1)
-#endif
-
-
-
-#ifndef S_ISUID
-# define S_ISUID 0004000
-#endif
-#ifndef S_ISGID
-# define S_ISGID 0002000
-#endif
-#ifndef S_ISVTX
-# define S_ISVTX 0001000
-#endif
-#ifndef S_IRUSR
-# define S_IRUSR 0000400
-#endif
-#ifndef S_IWUSR
-# define S_IWUSR 0000200
-#endif
-#ifndef S_IXUSR
-# define S_IXUSR 0000100
-#endif
-#ifndef S_IRGRP
-# define S_IRGRP 0000040
-#endif
-#ifndef S_IWGRP
-# define S_IWGRP 0000020
-#endif
-#ifndef S_IXGRP
-# define S_IXGRP 0000010
-#endif
-#ifndef S_IROTH
-# define S_IROTH 0000004
-#endif
-#ifndef S_IWOTH
-# define S_IWOTH 0000002
-#endif
-#ifndef S_IXOTH
-# define S_IXOTH 0000001
-#endif
 
 #define MODE_WXUSR	(S_IWUSR | S_IXUSR)
 #define MODE_R		(S_IRUSR | S_IRGRP | S_IROTH)
@@ -140,21 +76,12 @@ int get_statinfo PARAMS((const char *pathname, const char *name, struct stat *p)
 #define MODE_RWX	(S_IXUSR | S_IXGRP | S_IXOTH | MODE_RW)
 #define MODE_ALL	(S_ISUID | S_ISGID | S_ISVTX | MODE_RWX)
 
-#if 1
-#include <stdbool.h>
-typedef bool boolean;
-#else
-/* Not char because of type promotion; NeXT gcc can't handle it.  */
-typedef int boolean;
-#define		true    1
-#define		false	0
-#endif
 
 struct predicate;
 struct options;
 
 /* Pointer to a predicate function. */
-typedef boolean (*PRED_FUNC)(char *pathname, struct stat *stat_buf, struct predicate *pred_ptr);
+typedef boolean (*PRED_FUNC)(const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr);
 
 /* The number of seconds in a day. */
 #define		DAYSECS	    86400
@@ -208,8 +135,7 @@ struct perm_val
   mode_t val[2];
 };
 
-/* dir_id is used to support loop detection in find.c and 
- * also to support the -samefile test.
+/* dir_id is used to support loop detection in find.c
  */
 struct dir_id
 {
@@ -217,6 +143,15 @@ struct dir_id
   dev_t dev;
 };
 
+/* samefile_file_id is used to support the -samefile test.
+ */
+struct samefile_file_id
+{
+  ino_t ino;
+  dev_t dev;
+  int   fd;
+};
+
 struct size_val
 {
   enum comparison_type kind;
@@ -280,6 +215,13 @@ struct format_val
   struct quoting_options *quote_opts;
 };
 
+/* Profiling information for a predicate */
+struct predicate_performance_info
+{
+  unsigned long visits;
+  unsigned long successes;
+};
+
 /* evaluation cost of a predicate */
 enum EvaluationCost
 {
@@ -354,7 +296,7 @@ struct predicate
     gid_t gid;			/* group */
     struct time_val reftime;	/* newer newerXY anewer cnewer mtime atime ctime mmin amin cmin */
     struct perm_val perm;	/* perm */
-    struct dir_id   fileid;	/* samefile */
+    struct samefile_file_id samefileid; /* samefile */
     mode_t type;		/* type */
     FILE *stream;		/* ls fls fprint0 */
     struct format_val printf_vec; /* printf fprintf fprint  */
@@ -371,6 +313,8 @@ struct predicate
   struct predicate *pred_left;
   struct predicate *pred_right;
 
+  struct predicate_performance_info perf;
+  
   const struct parser_table* parser_entry;
 };
 
@@ -381,27 +325,6 @@ int get_current_dirfd(void);
 
 /* find library function declarations.  */
 
-/* dirname.c */
-char *dirname PARAMS((char *path));
-
-/* error.c */
-void error PARAMS((int status, int errnum, char *message, ...));
-
-/* stpcpy.c */
-#if !HAVE_STPCPY
-char *stpcpy PARAMS((char *dest, const char *src));
-#endif
-
-/* xgetcwd.c */
-char *xgetcwd PARAMS((void));
-
-/* xmalloc.c */
-#if __STDC__
-#define VOID void
-#else
-#define VOID char
-#endif
-
 /* find global function declarations.  */
 
 /* find.c */
@@ -452,8 +375,6 @@ struct parser_table
 
 /* parser.c */
 const struct parser_table* find_parser PARAMS((char *search_name));
-boolean parse_open  PARAMS((const struct parser_table* entry, char *argv[], int *arg_ptr));
-boolean parse_close PARAMS((const struct parser_table* entry, char *argv[], int *arg_ptr));
 boolean parse_print PARAMS((const struct parser_table*, char *argv[], int *arg_ptr));
 void pred_sanity_check PARAMS((const struct predicate *predicates));
 void parse_begin_user_args PARAMS((char **args, int argno, const struct predicate *last, const struct predicate *predicates));
@@ -462,64 +383,66 @@ boolean parse_openparen              PARAMS((const struct parser_table* entry, c
 boolean parse_closeparen             PARAMS((const struct parser_table* entry, char *argv[], int *arg_ptr));
 
 /* pred.c */
-boolean pred_amin PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_and PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_anewer PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_atime PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_close PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_cmin PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_cnewer PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_comma PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_ctime PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_delete PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_empty PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_exec PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_execdir PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_executable PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_false PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_fls PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_fprint PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_fprint0 PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_fprintf PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_fstype PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_gid PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_group PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_ilname PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_iname PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_inum PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_ipath PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_links PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_lname PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_ls PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_mmin PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_mtime PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_name PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_negate PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_newer PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_newerXY PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_nogroup PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_nouser PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_ok PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_okdir PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_open PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_or PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_path PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_perm PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_print PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_print0 PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_prune PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_quit PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_readable PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_regex PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_samefile PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_size PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_true PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_type PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_uid PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_used PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_user PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_writable PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_xtype PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
+
+typedef boolean PREDICATEFUNCTION(const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr);
+PREDICATEFUNCTION pred_amin;
+PREDICATEFUNCTION pred_and;
+PREDICATEFUNCTION pred_anewer;
+PREDICATEFUNCTION pred_atime;
+PREDICATEFUNCTION pred_closeparen;
+PREDICATEFUNCTION pred_cmin;
+PREDICATEFUNCTION pred_cnewer;
+PREDICATEFUNCTION pred_comma;
+PREDICATEFUNCTION pred_ctime;
+PREDICATEFUNCTION pred_delete;
+PREDICATEFUNCTION pred_empty;
+PREDICATEFUNCTION pred_exec;
+PREDICATEFUNCTION pred_execdir;
+PREDICATEFUNCTION pred_executable;
+PREDICATEFUNCTION pred_false;
+PREDICATEFUNCTION pred_fls;
+PREDICATEFUNCTION pred_fprint;
+PREDICATEFUNCTION pred_fprint0;
+PREDICATEFUNCTION pred_fprintf;
+PREDICATEFUNCTION pred_fstype;
+PREDICATEFUNCTION pred_gid;
+PREDICATEFUNCTION pred_group;
+PREDICATEFUNCTION pred_ilname;
+PREDICATEFUNCTION pred_iname;
+PREDICATEFUNCTION pred_inum;
+PREDICATEFUNCTION pred_ipath;
+PREDICATEFUNCTION pred_links;
+PREDICATEFUNCTION pred_lname;
+PREDICATEFUNCTION pred_ls;
+PREDICATEFUNCTION pred_mmin;
+PREDICATEFUNCTION pred_mtime;
+PREDICATEFUNCTION pred_name;
+PREDICATEFUNCTION pred_negate;
+PREDICATEFUNCTION pred_newer;
+PREDICATEFUNCTION pred_newerXY;
+PREDICATEFUNCTION pred_nogroup;
+PREDICATEFUNCTION pred_nouser;
+PREDICATEFUNCTION pred_ok;
+PREDICATEFUNCTION pred_okdir;
+PREDICATEFUNCTION pred_openparen;
+PREDICATEFUNCTION pred_or;
+PREDICATEFUNCTION pred_path;
+PREDICATEFUNCTION pred_perm;
+PREDICATEFUNCTION pred_print;
+PREDICATEFUNCTION pred_print0;
+PREDICATEFUNCTION pred_prune;
+PREDICATEFUNCTION pred_quit;
+PREDICATEFUNCTION pred_readable;
+PREDICATEFUNCTION pred_regex;
+PREDICATEFUNCTION pred_samefile;
+PREDICATEFUNCTION pred_size;
+PREDICATEFUNCTION pred_true;
+PREDICATEFUNCTION pred_type;
+PREDICATEFUNCTION pred_uid;
+PREDICATEFUNCTION pred_used;
+PREDICATEFUNCTION pred_user;
+PREDICATEFUNCTION pred_writable;
+PREDICATEFUNCTION pred_xtype;
 
 
 
@@ -535,6 +458,7 @@ void print_predicate PARAMS((FILE *fp, const struct predicate *p));
 void print_tree PARAMS((FILE*, struct predicate *node, int indent));
 void print_list PARAMS((FILE*, struct predicate *node));
 void print_optlist PARAMS((FILE *fp, const struct predicate *node));
+void show_success_rates(const struct predicate *node);
 
 
 /* tree.c */
@@ -551,10 +475,21 @@ void usage PARAMS((FILE *fp, int status, char *msg));
 extern boolean check_nofollow(void);
 void complete_pending_execs(struct predicate *p);
 void complete_pending_execdirs(int dirfd); /* Passing dirfd is an unpleasant CodeSmell. */
+const char *safely_quote_err_filename (int n, char const *arg);
+void fatal_file_error(const char *name) ATTRIBUTE_NORETURN;
 
 int process_leading_options PARAMS((int argc, char *argv[]));
 void set_option_defaults PARAMS((struct options *p));
 
+#if 0
+#define apply_predicate(pathname, stat_buf_ptr, node)	\
+  (*(node)->pred_func)((pathname), (stat_buf_ptr), (node))
+#else
+boolean apply_predicate(const char *pathname, struct stat *stat_buf, struct predicate *p);
+#endif
+
+#define pred_is(node, fn) ( ((node)->pred_func) == (fn) )
+
 
 /* find.c. */
 int get_info PARAMS((const char *pathname, struct stat *p, struct predicate *pred_ptr));
@@ -572,7 +507,8 @@ enum DebugOption
     DebugSearch           = 4,
     DebugTreeOpt          = 8,
     DebugHelp             = 16,
-    DebugExec             = 32
+    DebugExec             = 32,
+    DebugSuccessRates     = 64
   };
 
 struct options
@@ -642,6 +578,11 @@ struct options
   /* Optimisation level.  One is the default. 
    */
   unsigned short optimisation_level;
+
+
+  /* How should we quote filenames in error messages and so forth?
+   */
+  enum quoting_style err_quoting_style;
 };
 extern struct options options;
 
@@ -690,5 +631,4 @@ extern int starting_desc;
 extern char *program_name;
 
 
-
 #endif
