@@ -57,6 +57,9 @@
 #include "fts_.h"
 #include "openat.h"
 #include "save-cwd.h"
+#include "xgetcwd.h"
+#include "error.h"
+#include "dircallback.h"
 
 #ifdef HAVE_LOCALE_H
 #include <locale.h>
@@ -237,7 +240,7 @@ visit(FTS *p, FTSENT *ent, struct stat *pstat)
 
   /* Apply the predicates to this path. */
   eval_tree = get_eval_tree();
-  (*(eval_tree)->pred_func)(ent->fts_path, pstat, eval_tree);
+  apply_predicate(ent->fts_path, pstat, eval_tree);
 
   /* Deal with any side effects of applying the predicates. */
   if (state.stop_at_current_level)
@@ -285,7 +288,7 @@ issue_loop_warning(FTSENT * ent)
     {
       error(0, 0,
 	    _("Symbolic link %s is part of a loop in the directory hierarchy; we have already visited the directory to which it points."),
-	    quotearg_n_style(0, locale_quoting_style, ent->fts_path));
+	    safely_quote_err_filename(0, ent->fts_path));
     }
   else
     {
@@ -300,11 +303,11 @@ issue_loop_warning(FTSENT * ent)
       error(0, 0,
 	    _("Filesystem loop detected; "
 	      "%s is part of the same filesystem loop as %s."),
-	    quotearg_n_style(0, locale_quoting_style, ent->fts_path),
+	    safely_quote_err_filename(0, ent->fts_path),
 	    partial_quotearg_n(1,
 			       ent->fts_cycle->fts_path,
 			       ent->fts_cycle->fts_pathlen,
-			       locale_quoting_style));
+			       options.err_quoting_style));
     }
 }
 
@@ -333,9 +336,10 @@ complete_execdirs_cb(void *context)
   (void) context;
   /* By the tme this callback is called, the current directory is correct. */
   complete_pending_execdirs(AT_FDCWD);
+  return 0;
 }
 
-static int 
+static void
 show_outstanding_execdirs(FILE *fp)
 {
   if (options.debug_options & DebugExec)
@@ -349,9 +353,9 @@ show_outstanding_execdirs(FILE *fp)
 	{
 	  const char *pfx;
 	  
-	  if (p->pred_func == pred_execdir)
+	  if (pred_is(p, pred_execdir))
 	    pfx = "-execdir";
-	  else if (p->pred_func == pred_okdir)
+	  else if (pred_is(p, pred_okdir))
 	    pfx = "-okdir";
 	  else
 	    pfx = NULL;
@@ -391,7 +395,6 @@ consider_visiting(FTS *p, FTSENT *ent)
   struct stat statbuf;
   mode_t mode;
   int ignore, isdir;
-  boolean dirchanged;
   
   if (options.debug_options & DebugSearch)
     fprintf(stderr,
@@ -399,8 +402,8 @@ consider_visiting(FTS *p, FTSENT *ent)
             "fts_path=%s, fts_accpath=%s\n",
 	    get_fts_info_name(ent->fts_info),
             (int)ent->fts_level, prev_depth,
-	    quotearg_n_style(0, locale_quoting_style, ent->fts_path),
-	    quotearg_n_style(1, locale_quoting_style, ent->fts_accpath));
+	    quotearg_n_style(0, options.err_quoting_style, ent->fts_path),
+	    quotearg_n_style(1, options.err_quoting_style, ent->fts_accpath));
   
   if (ent->fts_info == FTS_DP)
     {
@@ -418,7 +421,8 @@ consider_visiting(FTS *p, FTSENT *ent)
   if (ent->fts_info == FTS_ERR
       || ent->fts_info == FTS_DNR)
     {
-      error(0, ent->fts_errno, ent->fts_path);
+      error(0, ent->fts_errno, "%s",
+	    safely_quote_err_filename(0, ent->fts_path));
       error_severity(1);
       return;
     }
@@ -439,7 +443,7 @@ consider_visiting(FTS *p, FTSENT *ent)
        */
       if (symlink_loop(ent->fts_accpath))
 	{
-	  error(0, ELOOP, ent->fts_path);
+	  error(0, ELOOP, "%s", safely_quote_err_filename(0, ent->fts_path));
 	  error_severity(1);
 	  return;
 	}
@@ -449,7 +453,8 @@ consider_visiting(FTS *p, FTSENT *ent)
       if (ent->fts_level == 0)
 	{
 	  /* e.g., nonexistent starting point */
-	  error(0, ent->fts_errno, ent->fts_path);
+	  error(0, ent->fts_errno, "%s",
+		safely_quote_err_filename(0, ent->fts_path));
 	  error_severity(1);	/* remember problem */
 	  return;
 	}
@@ -460,7 +465,8 @@ consider_visiting(FTS *p, FTSENT *ent)
 	   */
 	  if (symlink_loop(ent->fts_accpath))
 	    {
-	      error(0, ELOOP, ent->fts_path);
+	      error(0, ELOOP, "%s",
+		    safely_quote_err_filename(0, ent->fts_path));
 	      error_severity(1);
 	      return;
 	    }
@@ -481,6 +487,13 @@ consider_visiting(FTS *p, FTSENT *ent)
       state.have_type = true;
       statbuf = *(ent->fts_statp);
       state.type = mode = statbuf.st_mode;
+      
+      if (00000 == mode)
+	{
+	  /* Savannah bug #16378. */
+	  error(0, 0, _("Warning: file %s appears to have mode 0000"),
+		quotearg_n_style(0, options.err_quoting_style, ent->fts_path));
+	}
     }
 
   if (mode)
@@ -553,7 +566,7 @@ consider_visiting(FTS *p, FTSENT *ent)
   if (state.execdirs_outstanding)
     {
       show_outstanding_execdirs(stderr);
-      run_in_dir(p->fts_cwd_fd, complete_execdirs_cb);
+      run_in_dir(p->fts_cwd_fd, complete_execdirs_cb, NULL);
     }
 
   if (ent->fts_info == FTS_DP)
@@ -600,9 +613,8 @@ find(char *arg)
   p = fts_open(arglist, ftsoptions, NULL);
   if (NULL == p)
     {
-      error (0, errno,
-	     _("cannot search %s"),
-	     quotearg_n_style(0, locale_quoting_style, arg));
+      error (0, errno, _("cannot search %s"),
+	     safely_quote_err_filename(0, arg));
     }
   else
     {
@@ -732,6 +744,7 @@ main (int argc, char **argv)
    * partially-full command lines which have been built, 
    * but which are not yet complete.   Execute those now.
    */
+  show_success_rates(eval_tree);
   cleanup();
   return state.exit_status;
 }
