@@ -39,6 +39,8 @@
 #include "stat-time.h"
 #include "xstrtod.h"
 #include "fts_.h"
+#include "getdate.h"
+#include "error.h"
 #include "gnulib-version.h"
 
 #ifdef HAVE_FCNTL_H
@@ -201,7 +203,11 @@ static boolean parse_noop PARAMS((const struct parser_table* entry, char **argv,
   { (ARG_PUNCTUATION), (what), PASTE(parse_,suffix), PASTE(pred_,suffix) }
 
 
-/* Predicates we cannot handle in the usual way */
+/* Predicates we cannot handle in the usual way.  If you add an entry
+ * to this table, double-check the switch statement in
+ * pred_sanity_check() to make sure that the new case is being
+ * correctly handled.
+ */
 static struct parser_table const parse_entry_newerXY = 
   {
     ARG_SPECIAL_PARSE, "newerXY",            parse_newerXY, pred_newerXY /* BSD  */
@@ -214,8 +220,8 @@ static struct parser_table const parse_table[] =
 {
   PARSE_PUNCTUATION("!",                     negate),
   PARSE_PUNCTUATION("not",                   negate),	     /* GNU */
-  PARSE_PUNCTUATION("(",                     open),
-  PARSE_PUNCTUATION(")",                     close),
+  PARSE_PUNCTUATION("(",                     openparen),
+  PARSE_PUNCTUATION(")",                     closeparen),
   PARSE_PUNCTUATION(",",                     comma),	     /* GNU */
   PARSE_PUNCTUATION("a",                     and),
   PARSE_TEST       ("amin",                  amin),	     /* GNU */
@@ -490,7 +496,6 @@ const struct parser_table*
 find_parser (char *search_name)
 {
   int i;
-  const struct parser_table *p;
   const char *original_arg = search_name;
   
   /* Ugh.  Special case -newerXY. */
@@ -582,7 +587,7 @@ parse_anewer (const struct parser_table* entry, char **argv, int *arg_ptr)
     return false;
   set_stat_placeholders(&stat_newer);
   if ((*options.xstat) (argv[*arg_ptr], &stat_newer))
-    error (1, errno, "%s", argv[*arg_ptr]);
+    fatal_file_error(argv[*arg_ptr]);
   our_pred = insert_primary (entry);
   our_pred->args.reftime.xval = XVAL_ATIME;
   our_pred->args.reftime.ts = get_stat_mtime(&stat_newer);
@@ -593,7 +598,7 @@ parse_anewer (const struct parser_table* entry, char **argv, int *arg_ptr)
 }
 
 boolean
-parse_close (const struct parser_table* entry, char **argv, int *arg_ptr)
+parse_closeparen (const struct parser_table* entry, char **argv, int *arg_ptr)
 {
   struct predicate *our_pred;
 
@@ -601,7 +606,7 @@ parse_close (const struct parser_table* entry, char **argv, int *arg_ptr)
   (void) arg_ptr;
   
   our_pred = get_new_pred (entry);
-  our_pred->pred_func = pred_close;
+  our_pred->pred_func = pred_closeparen;
   our_pred->p_type = CLOSE_PAREN;
   our_pred->p_prec = NO_PREC;
   our_pred->need_stat = our_pred->need_type = false;
@@ -618,7 +623,7 @@ parse_cnewer (const struct parser_table* entry, char **argv, int *arg_ptr)
     return false;
   set_stat_placeholders(&stat_newer);
   if ((*options.xstat) (argv[*arg_ptr], &stat_newer))
-    error (1, errno, "%s", argv[*arg_ptr]);
+    fatal_file_error(argv[*arg_ptr]);
   our_pred = insert_primary (entry);
   our_pred->args.reftime.xval = XVAL_CTIME; /* like -newercm */
   our_pred->args.reftime.ts = get_stat_mtime(&stat_newer);
@@ -673,11 +678,17 @@ parse_delete (const struct parser_table* entry, char *argv[], int *arg_ptr)
   struct predicate *our_pred;
   (void) argv;
   (void) arg_ptr;
-  
+
   our_pred = insert_primary (entry);
   our_pred->side_effects = our_pred->no_default_print = true;
   /* -delete implies -depth */
   options.do_dir_first = false;
+  
+  /* We do not need stat information because we check for the case
+   * (errno==EISDIR) in pred_delete.
+   */
+  our_pred->need_stat = our_pred->need_type = false;
+  
   our_pred->est_success_rate = 1.0f;
   return true;
 }
@@ -998,7 +1009,9 @@ check_name_arg(const char *pred, const char *arg)
   if (strchr(arg, '/'))
     {
       error(0, 0,_("warning: Unix filenames usually don't contain slashes (though pathnames do).  That means that '%s %s' will probably evaluate to false all the time on this system.  You might find the '-wholename' test more useful, or perhaps '-samefile'.  Alternatively, if you are using GNU grep, you could use 'find ... -print0 | grep -FzZ %s'."),
-	    pred, arg, arg);
+	    pred,
+	    safely_quote_err_filename(0, arg),
+	    safely_quote_err_filename(1, arg));
     }
   return true;			/* allow it anyway */
 }
@@ -1246,7 +1259,7 @@ parse_newer (const struct parser_table* entry, char **argv, int *arg_ptr)
     return false;
   set_stat_placeholders(&stat_newer);
   if ((*options.xstat) (argv[*arg_ptr], &stat_newer))
-    error (1, errno, "%s", argv[*arg_ptr]);
+    fatal_file_error(argv[*arg_ptr]);
   our_pred = insert_primary (entry);
   our_pred->args.reftime.ts = get_stat_mtime(&stat_newer);
   our_pred->args.reftime.xval = XVAL_MTIME;
@@ -1335,8 +1348,8 @@ parse_newerXY (const struct parser_table* entry, char **argv, int *arg_ptr)
 			    &options.start_time))
 		{
 		  error(1, 0,
-			_("I cannot figure out how to interpret `%s' as a date or time"),
-			argv[*arg_ptr]);
+			_("I cannot figure out how to interpret %s as a date or time"),
+			quotearg_n_style(0, options.err_quoting_style, argv[*arg_ptr]));
 		}
 	    }
 	  else
@@ -1346,13 +1359,13 @@ parse_newerXY (const struct parser_table* entry, char **argv, int *arg_ptr)
 	      /* Stat the named file. */
 	      set_stat_placeholders(&stat_newer);
 	      if ((*options.xstat) (argv[*arg_ptr], &stat_newer))
-		error (1, errno, "%s", argv[*arg_ptr]);
+		fatal_file_error(argv[*arg_ptr]);
 	      
 	      if (!get_stat_Ytime(&stat_newer, y, &our_pred->args.reftime.ts))
 		{
 		  /* We cannot extract a timestamp from the struct stat. */
-		  error(1, 0, _("Cannot obtain birth time of file `%s'"),
-			argv[*arg_ptr]);
+		  error(1, 0, _("Cannot obtain birth time of file %s"),
+			safely_quote_err_filename(0, argv[*arg_ptr]));
 		}
 	    }
 	  our_pred->args.reftime.kind = COMP_GT;
@@ -1493,7 +1506,7 @@ parse_okdir (const struct parser_table* entry, char **argv, int *arg_ptr)
 }
 
 boolean
-parse_open (const struct parser_table* entry, char **argv, int *arg_ptr)
+parse_openparen (const struct parser_table* entry, char **argv, int *arg_ptr)
 {
   struct predicate *our_pred;
 
@@ -1501,7 +1514,7 @@ parse_open (const struct parser_table* entry, char **argv, int *arg_ptr)
   (void) arg_ptr;
   
   our_pred = get_new_pred_chk_op (entry);
-  our_pred->pred_func = pred_open;
+  our_pred->pred_func = pred_openparen;
   our_pred->p_type = OPEN_PAREN;
   our_pred->p_prec = NO_PREC;
   our_pred->need_stat = our_pred->need_type = false;
@@ -1623,7 +1636,8 @@ parse_perm (const struct parser_table* entry, char **argv, int *arg_ptr)
     {
       change = mode_compile (argv[*arg_ptr] + mode_start);
       if (NULL == change)
-	error (1, 0, _("invalid mode `%s'"), argv[*arg_ptr]);
+	error (1, 0, _("invalid mode %s"),
+	       quotearg_n_style(0, options.err_quoting_style, argv[*arg_ptr]));
     }
   perm_val[0] = mode_adjust (0, false, 0, change, NULL);
   perm_val[1] = mode_adjust (0, true, 0, change, NULL);
@@ -1899,18 +1913,147 @@ parse_size (const struct parser_table* entry, char **argv, int *arg_ptr)
 static boolean
 parse_samefile (const struct parser_table* entry, char **argv, int *arg_ptr)
 {
+  /* General idea: stat the file, remember device and inode numbers.
+   * If a candidate file matches those, it's the same file.  
+   */
   struct predicate *our_pred;
-  struct stat st;
+  struct stat st, fst;
+  int fd, openflags;
   
   if ((argv == NULL) || (argv[*arg_ptr] == NULL))
     return false;
+
   set_stat_placeholders(&st);
+  set_stat_placeholders(&fst);
+  
   if ((*options.xstat) (argv[*arg_ptr], &st))
-    error (1, errno, "%s", argv[*arg_ptr]);
+    fatal_file_error(argv[*arg_ptr]);
+
+  /* POSIX systems are free to re-use the inode number of a deleted
+   * file.  To ensure that we are not fooled by inode reuse, we hold
+   * the file open if we can.  This would prevent the system reusing
+   * the file.
+   */
+  fd = -3;			/* means, uninitialised */
+  openflags = O_RDONLY;
+  
+  if (options.symlink_handling == SYMLINK_NEVER_DEREF)
+    {
+      if (options.open_nofollow_available)
+	{
+	  assert(O_NOFOLLOW != 0);
+	  openflags |= O_NOFOLLOW;
+	  fd = -1;		/* safe to open it. */
+	}
+      else
+	{
+	  if (S_ISLNK(st.st_mode))
+	    {
+	      /* no way to ensure that a symlink will not be followed
+	       * by open(2), so fall back on using lstat().  Accept 
+	       * the risk that the named file will be deleted and 
+	       * replaced with another having the same inode.
+	       *
+	       * Avoid opening the file.
+	       */
+	      fd = -2;		/* Do not open it */
+	    }
+	  else
+	    {
+	      fd = -1;
+	      /* Race condition here: the file might become a symlink here. */
+	    }
+	}
+    }
+  else
+    {
+      /* We want to dereference the symlink anyway */
+      fd = -1;			/* safe to open it without O_NOFOLLOW */
+    }
+
+  assert(fd != -3);		/* check we made a decision */
+  if (fd == -1)
+    {
+      /* Race condition here.  The file might become a
+       * symbolic link in between out call to stat and
+       * the call to open.
+       */
+      fd = open(argv[*arg_ptr], openflags);
+
+      if (fd >= 0)
+	{
+	  /* We stat the file again here to prevent a race condition 
+	   * between the first stat and the call to open(2).
+	   */
+	  if (0 != fstat(fd, &fst))
+	    {
+	      fatal_file_error(argv[*arg_ptr]);
+	    }
+	  else
+	    {
+	      /* Worry about the race condition.  If the file became a
+	       * symlink after our first stat and before our call to
+	       * open, fst may contain the stat information for the
+	       * destination of the link, not the link itself.
+	       */
+	      if ((*options.xstat) (argv[*arg_ptr], &st))
+		fatal_file_error(argv[*arg_ptr]);
+
+	      if ((options.symlink_handling == SYMLINK_NEVER_DEREF)
+		  && (!options.open_nofollow_available))
+		{
+		  if (S_ISLNK(st.st_mode))
+		    {
+		      /* We lost the race.  Leave the data in st.  The
+		       * file descriptor points to the wrong thing.
+		       */
+		      close(fd);
+		      fd = -1;
+		    }
+		  else
+		    {
+		      /* Several possibilities here:
+		       * 1. There was no race
+		       * 2. The file changed into a symlink after the stat and
+		       *    before the open, and then back into a non-symlink
+		       *    before the second stat.
+		       *
+		       * In case (1) there is no problem.  In case (2), 
+		       * the stat() and fstat() calls will have returned 
+		       * different data.  O_NOFOLLOW was not available,
+		       * so the open() call may have followed a symlink
+		       * even if the -P option is in effect.
+		       */
+		      if ((st.st_dev == fst.st_dev)
+			  && (st.st_ino == fst.st_ino))
+			{
+			  /* No race.  No need to copy fst to st,
+			   * since they should be identical (modulo
+			   * differences in padding bytes).
+			   */
+			}
+		      else
+			{
+			  /* We lost the race.  Leave the data in st.  The
+			   * file descriptor points to the wrong thing.
+			   */
+			  close(fd);
+			  fd = -1;
+			}
+		    }
+		}
+	      else
+		{
+		  st = fst;
+		}
+	    }
+	}
+    }
   
   our_pred = insert_primary (entry);
-  our_pred->args.fileid.ino = st.st_ino;
-  our_pred->args.fileid.dev = st.st_dev;
+  our_pred->args.samefileid.ino = st.st_ino;
+  our_pred->args.samefileid.dev = st.st_dev;
+  our_pred->args.samefileid.fd  = fd;
   our_pred->need_type = false;
   our_pred->need_stat = true;
   our_pred->est_success_rate = 0.01f;
@@ -1919,6 +2062,9 @@ parse_samefile (const struct parser_table* entry, char **argv, int *arg_ptr)
 }
 
 #if 0
+/* This function is commented out partly because support for it is
+ * uneven. 
+ */
 static boolean
 parse_show_control_chars (const struct parser_table* entry, char **argv, int *arg_ptr)
 {
@@ -1985,7 +2131,7 @@ parse_accesscheck (const struct parser_table* entry, char **argv, int *arg_ptr)
   our_pred = insert_primary (entry);
   our_pred->need_stat = our_pred->need_type = false;
   our_pred->side_effects = our_pred->no_default_print = false;
-  if (our_pred->pred_func == pred_executable)
+  if (pred_is(our_pred, pred_executable))
     our_pred->est_success_rate = 0.2;
   else
     our_pred->est_success_rate = 0.9;
@@ -2583,7 +2729,8 @@ check_path_safety(const char *action, char **argv)
 	{
 	  /* Relative paths are also dangerous in $PATH. */
 	  error(1, 0, _("The ralative path %s is included in the PATH environment variable, which is insecure in combination with the %s action of find.  Please remove that entry from $PATH"),
-		s, action);
+		safely_quote_err_filename(0, s),
+		action);
 	}
     }
 }
@@ -2665,7 +2812,7 @@ new_insert_exec_ok (const char *action,
 	}
       
       saw_braces = 0;
-      if (strstr (argv[end], "{}")) /* XXX: wrong for multibyte locales */
+      if (mbsstr (argv[end], "{}"))
 	{
 	  saw_braces = 1;
 	  ++brace_count;
@@ -2708,14 +2855,18 @@ new_insert_exec_ok (const char *action,
 	    suffix);
     }
 
-  /* We use a switch statement here so that 
-   * the compiler warns us when we forget to handle a 
-   * newly invented enum value.
+  /* We use a switch statement here so that the compiler warns us when
+   * we forget to handle a newly invented enum value.
+   *
+   * Like xargs, we allow 2KiB of headroom for the launched utility to
+   * export its own environment variables before calling something
+   * else.
    */
-  bcstatus = bc_init_controlinfo(&execp->ctl);
+  bcstatus = bc_init_controlinfo(&execp->ctl, 2048u);
   switch (bcstatus) 
     {
     case BC_INIT_ENV_TOO_BIG:
+    case BC_INIT_CANNOT_ACCOMODATE_HEADROOM:
       error(1, 0, 
 	    _("The environment is too large for exec()."));
       break;
@@ -3052,6 +3203,6 @@ open_output_file (char *path)
     return stdout;
   f = fopen_safer (path, "w");
   if (f == NULL)
-    error (1, errno, "%s", path);
+    fatal_file_error(path);
   return f;
 }
