@@ -43,6 +43,7 @@
 #include "listfile.h"
 #include "stat-time.h"
 #include "dircallback.h"
+#include "error.h"
 
 #if ENABLE_NLS
 # include <libintl.h>
@@ -155,7 +156,7 @@
 #undef MAX
 #define MAX(a, b) ((a) > (b) ? (a) : (b))
 
-static boolean match_lname PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr, boolean ignore_case));
+static boolean match_lname PARAMS((const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr, boolean ignore_case));
 
 static char *format_date PARAMS((struct timespec ts, int kind));
 static char *ctime_format PARAMS((struct timespec ts));
@@ -173,7 +174,7 @@ struct pred_assoc pred_table[] =
   {pred_and, "and     "},
   {pred_anewer, "anewer  "},
   {pred_atime, "atime   "},
-  {pred_close, ")       "},
+  {pred_closeparen, ")       "},
   {pred_cmin, "cmin    "},
   {pred_cnewer, "cnewer  "},
   {pred_comma, ",       "},
@@ -207,7 +208,7 @@ struct pred_assoc pred_table[] =
   {pred_nouser, "nouser  "},
   {pred_ok, "ok      "},
   {pred_okdir, "okdir   "},
-  {pred_open, "(       "},
+  {pred_openparen, "(       "},
   {pred_or, "or      "},
   {pred_path, "path    "},
   {pred_perm, "perm    "},
@@ -275,8 +276,6 @@ compare_ts(struct timespec ts1,
 static boolean
 pred_timewindow(struct timespec ts, struct predicate const *pred_ptr, int window)
 {
-  double delta;
-  
   switch (pred_ptr->args.reftime.kind)
     {
     case COMP_GT:
@@ -286,38 +285,35 @@ pred_timewindow(struct timespec ts, struct predicate const *pred_ptr, int window
       return compare_ts(ts, pred_ptr->args.reftime.ts) < 0;
       
     case COMP_EQ:
-      delta = ts_difference(ts, pred_ptr->args.reftime.ts);
-      return (delta >= 0.0 && delta < window);
+      {
+	double delta = ts_difference(ts, pred_ptr->args.reftime.ts);
+	return (delta >= 0.0 && delta < window);
+      }
     }
 }
 
 
 boolean
-pred_amin (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_amin (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) &pathname;
   return pred_timewindow(get_stat_atime(stat_buf), pred_ptr, 60);
 }
 
 boolean
-pred_and (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_and (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   if (pred_ptr->pred_left == NULL
-      || (*pred_ptr->pred_left->pred_func) (pathname, stat_buf,
-					    pred_ptr->pred_left))
+      || apply_predicate(pathname, stat_buf, pred_ptr->pred_left))
     {
-      /* Check whether we need a stat here. */
-      if (get_info(pathname, stat_buf, pred_ptr) != 0)
-	    return false;
-      return ((*pred_ptr->pred_right->pred_func) (pathname, stat_buf,  
-						  pred_ptr->pred_right));
+      return apply_predicate(pathname, stat_buf, pred_ptr->pred_right);
     }
   else
-    return (false);
+    return false;
 }
 
 boolean
-pred_anewer (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_anewer (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) &pathname;
   assert(COMP_GT == pred_ptr->args.reftime.kind);
@@ -325,14 +321,14 @@ pred_anewer (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_atime (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_atime (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) &pathname;
   return pred_timewindow(get_stat_atime(stat_buf), pred_ptr, DAYSECS);
 }
 
 boolean
-pred_close (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_closeparen (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) &pathname;
   (void) &stat_buf;
@@ -342,14 +338,14 @@ pred_close (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_cmin (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_cmin (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) pathname;
   return pred_timewindow(get_stat_ctime(stat_buf), pred_ptr, 60);
 }
 
 boolean
-pred_cnewer (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_cnewer (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) pathname;
   
@@ -358,53 +354,69 @@ pred_cnewer (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_comma (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_comma (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   if (pred_ptr->pred_left != NULL)
-    (*pred_ptr->pred_left->pred_func) (pathname, stat_buf,
-				       pred_ptr->pred_left);
-  /* Check whether we need a stat here. */
-  /* TODO: what about need_type? */
-  if (get_info(pathname, stat_buf, pred_ptr) != 0)
-    return false;
-  return ((*pred_ptr->pred_right->pred_func) (pathname, stat_buf,
-					      pred_ptr->pred_right));
+    {
+      apply_predicate(pathname, stat_buf,pred_ptr->pred_left);
+    }
+  return apply_predicate(pathname, stat_buf, pred_ptr->pred_right);
 }
 
 boolean
-pred_ctime (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_ctime (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) &pathname;
   return pred_timewindow(get_stat_ctime(stat_buf), pred_ptr, DAYSECS);
 }
 
+static boolean
+perform_delete(int flags)
+{
+  return 0 == unlinkat(state.cwd_dir_fd, state.rel_pathname, flags);
+}
+
+
 boolean
-pred_delete (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_delete (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) pred_ptr;
   (void) stat_buf;
   if (strcmp (state.rel_pathname, "."))
     {
       int flags=0;
-      if (S_ISDIR(stat_buf->st_mode))
+      if (state.have_stat && S_ISDIR(stat_buf->st_mode))
 	flags |= AT_REMOVEDIR;
-      if (0 != unlinkat(state.cwd_dir_fd, state.rel_pathname, flags))
-	{
-	  error (0, errno, "cannot delete %s", pathname);
-	  return false;
-	}
-      else
+      if (perform_delete(flags))
 	{
 	  return true;
 	}
+      else
+	{
+	  if (EISDIR == errno)
+	    {
+	      if ((flags & AT_REMOVEDIR) == 0)
+		{
+		  /* unlink() operation failed because we should have done rmdir(). */
+		  flags |= AT_REMOVEDIR;
+		  if (perform_delete(flags))
+		    return true;
+		}
+	    }
+	}
+      error (0, errno, "cannot delete %s",
+	     safely_quote_err_filename(0, pathname));
+      return false;
+    }
+  else
+    {
+      /* nothing to do. */
+      return true;
     }
-  
-  /* nothing to do. */
-  return true;
 }
 
 boolean
-pred_empty (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_empty (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) pathname;
   (void) pred_ptr;
@@ -423,14 +435,14 @@ pred_empty (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 #endif
 		       )) < 0)
 	{
-	  error (0, errno, "%s", pathname);
+	  error (0, errno, "%s", safely_quote_err_filename(0, pathname));
 	  state.exit_status = 1;
 	  return false;
 	}
       d = fdopendir (fd);
       if (d == NULL)
 	{
-	  error (0, errno, "%s", pathname);
+	  error (0, errno, "%s", safely_quote_err_filename(0, pathname));
 	  state.exit_status = 1;
 	  return false;
 	}
@@ -446,7 +458,7 @@ pred_empty (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 	}
       if (CLOSEDIR (d))
 	{
-	  error (0, errno, "%s", pathname);
+	  error (0, errno, "%s", safely_quote_err_filename(0, pathname));
 	  state.exit_status = 1;
 	  return false;
 	}
@@ -512,14 +524,14 @@ new_impl_pred_exec (int dirfd, const char *pathname,
 
 
 boolean
-pred_exec (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_exec (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   return new_impl_pred_exec(get_start_dirfd(),
 			    pathname, stat_buf, pred_ptr, NULL, 0);
 }
 
 boolean
-pred_execdir (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_execdir (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
    const char *prefix = (state.rel_pathname[0] == '/') ? NULL : "./";
    (void) &pathname;
@@ -529,7 +541,7 @@ pred_execdir (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_false (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_false (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) &pathname;
   (void) &stat_buf;
@@ -540,7 +552,7 @@ pred_false (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_fls (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_fls (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   list_file (pathname, state.cwd_dir_fd, state.rel_pathname, stat_buf, options.start_time.tv_sec,
 	     options.output_block_size,
@@ -549,7 +561,7 @@ pred_fls (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_fprint (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_fprint (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) &pathname;
   (void) &stat_buf;
@@ -563,7 +575,7 @@ pred_fprint (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_fprint0 (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_fprint0 (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) &pathname;
   (void) &stat_buf;
@@ -613,13 +625,13 @@ file_sparseness(const struct stat *p)
 static boolean
 do_fprintf(FILE *fp,
 	   struct segment *segment,
-	   char *pathname,
+	   const char *pathname,
 	   const struct stat *stat_buf,
 	   boolean ttyflag,
 	   const struct quoting_options *qopts)
 {
   char hbuf[LONGEST_HUMAN_READABLE + 1];
-  char *cp;
+  const char *cp;
 
   switch (segment->segkind)
     {
@@ -700,8 +712,6 @@ do_fprintf(FILE *fp,
 	case 'h':		/* leading directories part of path */
 	  /* sanitised */
 	  {
-	    char cc;
-	    
 	    cp = strrchr (pathname, '/');
 	    if (cp == NULL)	/* No leading directories. */
 	      {
@@ -713,21 +723,21 @@ do_fprintf(FILE *fp,
 	      }
 	    else
 	      {
-		cc = *cp;
-		*cp = '\0';
-		print_quoted (fp, qopts, ttyflag, segment->text, pathname);
-		*cp = cc;
+		char *s = strdup(pathname);
+		s[cp - pathname] = 0;
+		print_quoted (fp, qopts, ttyflag, segment->text, s);
+		free(s);
 	      }
 	    break;
 	  }
 	case 'H':		/* ARGV element file was found under */
 	  /* trusted */
 	  {
-	    char cc = pathname[state.starting_path_length];
-
-	    pathname[state.starting_path_length] = '\0';
-	    fprintf (fp, segment->text, pathname);
-	    pathname[state.starting_path_length] = cc;
+	    char *s = xmalloc(state.starting_path_length+1);
+	    memcpy(s, pathname, state.starting_path_length);
+	    s[state.starting_path_length] = 0;
+	    fprintf (fp, segment->text, s);
+	    free(s);
 	    break;
 	  }
 	case 'i':		/* inode number */
@@ -906,7 +916,7 @@ do_fprintf(FILE *fp,
 		      fprintf (fp, segment->text, "L");
 		      break;
 		    };
-		    error (0, errno, "%s", pathname);
+		    error (0, errno, "%s", safely_quote_err_filename(0, pathname));
 		    /* exit_status = 1;
 		       return (false); */
 		  }
@@ -936,7 +946,7 @@ do_fprintf(FILE *fp,
 }
 
 boolean
-pred_fprintf (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_fprintf (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   FILE *fp = pred_ptr->args.printf_vec.stream;
   struct segment *segment;
@@ -1006,7 +1016,7 @@ pred_fprintf (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_fstype (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_fstype (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) pathname;
   
@@ -1017,7 +1027,7 @@ pred_fstype (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_gid (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_gid (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) pathname;
   
@@ -1040,7 +1050,7 @@ pred_gid (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_group (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_group (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) pathname;
   
@@ -1051,13 +1061,13 @@ pred_group (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_ilname (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_ilname (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   return match_lname (pathname, stat_buf, pred_ptr, true);
 }
 
 boolean
-pred_iname (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_iname (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   const char *base;
 
@@ -1073,7 +1083,7 @@ pred_iname (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_inum (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_inum (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) pathname;
   
@@ -1096,7 +1106,7 @@ pred_inum (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_ipath (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_ipath (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) stat_buf;
   
@@ -1106,7 +1116,7 @@ pred_ipath (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_links (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_links (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) pathname;
   
@@ -1129,13 +1139,13 @@ pred_links (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_lname (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_lname (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   return match_lname (pathname, stat_buf, pred_ptr, false);
 }
 
 static boolean
-match_lname (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr, boolean ignore_case)
+match_lname (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr, boolean ignore_case)
 {
   boolean ret = false;
 #ifdef S_ISLNK
@@ -1155,7 +1165,7 @@ match_lname (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr,
 }
 
 boolean
-pred_ls (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_ls (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   list_file (pathname, state.cwd_dir_fd, state.rel_pathname,
 	     stat_buf, options.start_time.tv_sec,
@@ -1166,21 +1176,21 @@ pred_ls (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_mmin (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_mmin (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) &pathname;
   return pred_timewindow(get_stat_mtime(stat_buf), pred_ptr, 60);
 }
 
 boolean
-pred_mtime (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_mtime (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) pathname;
   return pred_timewindow(get_stat_mtime(stat_buf), pred_ptr, DAYSECS);
 }
 
 boolean
-pred_name (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_name (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   const char *base;
 
@@ -1196,18 +1206,13 @@ pred_name (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_negate (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_negate (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
-  /* Check whether we need a stat here. */
-  /* TODO: what about need_type? */
-  if (get_info(pathname, stat_buf, pred_ptr) != 0)
-    return false;
-  return (!(*pred_ptr->pred_right->pred_func) (pathname, stat_buf,
-					      pred_ptr->pred_right));
+  return !apply_predicate(pathname, stat_buf, pred_ptr->pred_right);
 }
 
 boolean
-pred_newer (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_newer (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) pathname;
   
@@ -1216,43 +1221,53 @@ pred_newer (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_newerXY (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_newerXY (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   struct timespec ts;
+  boolean collected = false;
   
   assert(COMP_GT == pred_ptr->args.reftime.kind);
   
   switch (pred_ptr->args.reftime.xval)
     {
+    case XVAL_TIME:
+      assert(pred_ptr->args.reftime.xval != XVAL_TIME);
+      return false;
+
     case XVAL_ATIME:
       ts = get_stat_atime(stat_buf);
+      collected = true;
       break;
+      
     case XVAL_BIRTHTIME:
       ts = get_stat_birthtime(stat_buf);
+      collected = true;
       if (ts.tv_nsec < 0);
 	{
 	  /* XXX: Cannot determine birth time.  Warn once. */
-	  error(0, 0, _("Warning: cannot determine birth time of file `%s'"),
-		pathname);
-	  return 0;
+	  error(0, 0, _("Warning: cannot determine birth time of file %s"),
+		safely_quote_err_filename(0, pathname));
+	  return false;
 	}
       break;
+      
     case XVAL_CTIME:
       ts = get_stat_ctime(stat_buf);
+      collected = true;
       break;
+      
     case XVAL_MTIME:
       ts = get_stat_mtime(stat_buf);
-      break;
-    case XVAL_TIME:		/* Should not happen (-newertY is not valid). */
-      assert(pred_ptr->args.reftime.xval != XVAL_TIME);
-      abort();
+      collected = true;
       break;
     }
+  
+  assert(collected);
   return compare_ts(ts, pred_ptr->args.reftime.ts) > 0;
 }
 
 boolean
-pred_nogroup (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_nogroup (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) pathname;
   (void) pred_ptr;
@@ -1267,7 +1282,7 @@ pred_nogroup (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_nouser (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_nouser (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
 #ifdef CACHE_IDS
   extern char *uid_unused;
@@ -1300,7 +1315,7 @@ is_ok(const char *program, const char *arg)
 }
 
 boolean
-pred_ok (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_ok (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   if (is_ok(pred_ptr->args.exec_vec.replace_vec[0], pathname))
     return new_impl_pred_exec (get_start_dirfd(),
@@ -1310,7 +1325,7 @@ pred_ok (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_okdir (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_okdir (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   const char *prefix = (state.rel_pathname[0] == '/') ? NULL : "./";
   if (is_ok(pred_ptr->args.exec_vec.replace_vec[0], pathname))
@@ -1322,7 +1337,7 @@ pred_okdir (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_open (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_openparen (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) pathname;
   (void) stat_buf;
@@ -1331,23 +1346,19 @@ pred_open (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_or (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_or (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   if (pred_ptr->pred_left == NULL
-      || !(*pred_ptr->pred_left->pred_func) (pathname, stat_buf,
-					     pred_ptr->pred_left))
+      || !apply_predicate(pathname, stat_buf, pred_ptr->pred_left))
     {
-      if (get_info(pathname, stat_buf, pred_ptr) != 0)
-	return false;
-      return ((*pred_ptr->pred_right->pred_func) (pathname, stat_buf,
-						  pred_ptr->pred_right));
+      return apply_predicate(pathname, stat_buf, pred_ptr->pred_right);
     }
   else
     return true;
 }
 
 boolean
-pred_path (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_path (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) stat_buf;
   if (fnmatch (pred_ptr->args.str, pathname, 0) == 0)
@@ -1356,7 +1367,7 @@ pred_path (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_perm (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_perm (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   mode_t mode = stat_buf->st_mode;
   mode_t perm_val = pred_ptr->args.perm.val[S_ISDIR (mode) != 0];
@@ -1425,7 +1436,7 @@ can_access(int access_type)
 
 
 boolean
-pred_executable (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_executable (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) pathname;
   (void) stat_buf;
@@ -1435,7 +1446,7 @@ pred_executable (char *pathname, struct stat *stat_buf, struct predicate *pred_p
 }
 
 boolean
-pred_readable (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_readable (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) pathname;
   (void) stat_buf;
@@ -1445,7 +1456,7 @@ pred_readable (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr
 }
 
 boolean
-pred_writable (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_writable (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) pathname;
   (void) stat_buf;
@@ -1455,7 +1466,7 @@ pred_writable (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr
 }
 
 boolean
-pred_print (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_print (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) stat_buf;
   (void) pred_ptr;
@@ -1468,7 +1479,7 @@ pred_print (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_print0 (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_print0 (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) stat_buf;
   (void) pred_ptr;
@@ -1478,7 +1489,7 @@ pred_print0 (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_prune (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_prune (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) pathname;
   (void) pred_ptr;
@@ -1492,7 +1503,7 @@ pred_prune (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_quit (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_quit (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) pathname;
   (void) stat_buf;
@@ -1510,7 +1521,7 @@ pred_quit (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_regex (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_regex (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   int len = strlen (pathname);
 (void) stat_buf;
@@ -1521,7 +1532,7 @@ pred_regex (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_size (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_size (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   uintmax_t f_val;
 
@@ -1547,7 +1558,7 @@ pred_size (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_samefile (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_samefile (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   /* Potential optimisation: because of the loop protection, we always
    * know the device of the current directory, hence the device number
@@ -1558,13 +1569,17 @@ pred_samefile (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr
    * stat the file we're looking at.
    */
   (void) pathname;
-  
-  return stat_buf->st_ino == pred_ptr->args.fileid.ino
-    &&   stat_buf->st_dev == pred_ptr->args.fileid.dev;
+
+  /* We will often still have an fd open on the file under consideration,
+   * but that's just to ensure inode number stability by maintaining 
+   * a reference to it; we don't need the file for anything else.
+   */
+  return stat_buf->st_ino == pred_ptr->args.samefileid.ino
+    &&   stat_buf->st_dev == pred_ptr->args.samefileid.dev;
 }
 
 boolean
-pred_true (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_true (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) pathname;
   (void) stat_buf;
@@ -1573,13 +1588,20 @@ pred_true (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_type (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_type (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   mode_t mode;
   mode_t type = pred_ptr->args.type;
 
   assert(state.have_type);
-  assert(state.type != 0);
+
+  if (0 == state.type)
+    {
+      /* This can sometimes happen with broken NFS servers. 
+       * See Savannah bug #16378.
+       */
+      return false;
+    }
   
   (void) pathname;
 
@@ -1617,7 +1639,7 @@ pred_type (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_uid (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_uid (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) pathname;
   switch (pred_ptr->args.numinfo.kind)
@@ -1639,7 +1661,7 @@ pred_uid (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_used (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_used (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   struct timespec delta, at, ct;
 
@@ -1659,7 +1681,7 @@ pred_used (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_user (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_user (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) pathname;
   if (pred_ptr->args.uid == stat_buf->st_uid)
@@ -1669,7 +1691,7 @@ pred_user (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_xtype (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_xtype (const char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   struct stat sbuf;		/* local copy, not stat_buf because we're using a different stat method */
   int (*ystat) (const char*, struct stat *p);
@@ -1695,7 +1717,7 @@ pred_xtype (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 	}
       else
 	{
-	  error (0, errno, "%s", pathname);
+	  error (0, errno, "%s", safely_quote_err_filename(0, pathname));
 	  state.exit_status = 1;
 	}
       return false;
@@ -1751,7 +1773,7 @@ prep_child_for_exec (boolean close_stdin, int dirfd)
 	       * stdin is almost as good as executing it
 	       * with its stdin attached to /dev/null.
 	       */
-	      error (0, errno, "%s", inputfile);
+	      error (0, errno, "%s", safely_quote_err_filename(0, inputfile));
 	      /* do not set ok=false, it is OK to continue anyway. */
 	    }
 	}
@@ -1822,7 +1844,8 @@ launch (const struct buildcmd_control *ctl,
 	}
       
       execvp (buildstate->cmd_argv[0], buildstate->cmd_argv);
-      error (0, errno, "%s", buildstate->cmd_argv[0]);
+      error (0, errno, "%s",
+	     safely_quote_err_filename(0, buildstate->cmd_argv[0]));
       _exit (1);
     }
 
@@ -1835,7 +1858,8 @@ launch (const struct buildcmd_control *ctl,
     {
       if (errno != EINTR)
 	{
-	  error (0, errno, _("error waiting for %s"), buildstate->cmd_argv[0]);
+	  error (0, errno, _("error waiting for %s"),
+		 safely_quote_err_filename(0, buildstate->cmd_argv[0]));
 	  state.exit_status = 1;
 	  return 0;		/* FAIL */
 	}
@@ -1844,7 +1868,9 @@ launch (const struct buildcmd_control *ctl,
   if (WIFSIGNALED (wait_status))
     {
       error (0, 0, _("%s terminated by signal %d"),
-	     buildstate->cmd_argv[0], WTERMSIG (wait_status));
+	     quotearg_n_style(0, options.err_quoting_style,
+			      buildstate->cmd_argv[0]),
+	     WTERMSIG (wait_status));
       
       if (execp->multiple)
 	{
@@ -1926,6 +1952,9 @@ format_date (struct timespec ts, int kind)
    * measurable.
    */
   assert(sizeof(buf) >= LONGEST_HUMAN_READABLE);
+
+  charsprinted = 0;
+  need_ns_suffix = 0;
   
   /* Format the main part of the time. */
   if (kind == '+')
@@ -1956,9 +1985,10 @@ format_date (struct timespec ts, int kind)
 
   if (need_ns_suffix)
     {
-      /* Format the nanoseconds part.  Leave a trailing zero to discourage people from 
-       * writing scripts which extract the fractional part of the timestamp by using 
-       * column offsets.  The reason for discouraging this is that in the future, the 
+      /* Format the nanoseconds part.  Leave a trailing zero to
+       * discourage people from writing scripts which extract the
+       * fractional part of the timestamp by using column offsets.
+       * The reason for discouraging this is that in the future, the
        * granularity may not be nanoseconds.
        */
       ns_buf[0] = 0;
@@ -1970,7 +2000,9 @@ format_date (struct timespec ts, int kind)
       && (tm = localtime (&ts.tv_sec))
       && strftime (buf, sizeof buf, fmt, tm))
     {
-      /* For %AS, %CS, %TS, add the fractional part of the seconds information. */
+      /* For %AS, %CS, %TS, add the fractional part of the seconds
+       * information.
+       */
       if (need_ns_suffix)
 	{
 	  assert((sizeof buf - strlen(buf)) > strlen(ns_buf));
@@ -1993,9 +2025,10 @@ format_date (struct timespec ts, int kind)
       if (ts.tv_sec < 0)
 	*--p = '-'; /* XXX: Ugh, relying on internal details of human_readable(). */
 
-      /* Add the nanoseconds part.  Because we cannot enforce a particlar implementation 
-       * of human_readable, we cannot assume any particular value for (p-buf).  So we 
-       * need to be careful that there is enough space remaining in the buffer.
+      /* Add the nanoseconds part.  Because we cannot enforce a
+       * particlar implementation of human_readable, we cannot assume
+       * any particular value for (p-buf).  So we need to be careful
+       * that there is enough space remaining in the buffer.
        */
       if (need_ns_suffix)
 	{
@@ -2115,8 +2148,7 @@ print_parenthesised(FILE *fp, struct predicate *node)
 
   if (node)
     {
-      if ( ( (node->pred_func == pred_or)
-	     || (node->pred_func == pred_and) )
+      if ((pred_is(node, pred_or) || pred_is(node, pred_and))
 	  && node->pred_left == NULL)
 	{
 	  /* We print "<nothing> or  X" as just "X"
@@ -2150,9 +2182,35 @@ print_optlist (FILE *fp, const struct predicate *p)
 	       p->need_type ? "[need type] " : "");
       print_predicate(fp, p);
       fprintf(fp, " [%g] ", p->est_success_rate);
+      if (options.debug_options & DebugSuccessRates)
+	{
+	  fprintf(fp, "[%ld/%ld", p->perf.successes, p->perf.visits);
+	  if (p->perf.visits)
+	    {
+	      double real_rate = (double)p->perf.successes / (double)p->perf.visits;
+	      fprintf(fp, "=%g] ", real_rate);
+	    }
+	  else
+	    {
+	      fprintf(fp, "=_] ");
+	    }
+	}
       print_parenthesised(fp, p->pred_right);
     }
 }
+
+void show_success_rates(const struct predicate *p)
+{
+  if (options.debug_options & DebugSuccessRates)
+    {
+      fprintf(stderr, "Predicate success rates after completion:\n");
+      print_optlist(stderr, p);
+      fprintf(stderr, "\n");
+    }
+}
+
+
+
 
 #ifdef _NDEBUG
 /* If _NDEBUG is defined, the assertions will do nothing.   Hence 
@@ -2208,7 +2266,7 @@ pred_sanity_check(const struct predicate *predicates)
 	  
 	case ARG_ACTION:
 	  assert(p->side_effects); /* actions have side effects. */
-	  if (p->pred_func != pred_prune && p->pred_func != pred_quit)
+	  if (!pred_is(p, pred_prune) && !pred_is(p, pred_quit))
 	    {
 	      /* actions other than -prune and -quit should
 	       * inhibit the default -print
@@ -2217,8 +2275,12 @@ pred_sanity_check(const struct predicate *predicates)
 	    }
 	  break;
 
-	case ARG_PUNCTUATION:
+	/* We happen to know that the only user of ARG_SPECIAL_PARSE
+	 * is a test, so handle it like ARG_TEST.
+	 */
+	case ARG_SPECIAL_PARSE:
 	case ARG_TEST:
+	case ARG_PUNCTUATION:
 	case ARG_NOOP:
 	  /* Punctuation and tests should have no side
 	   * effects and not inhibit default print.
