@@ -1,5 +1,5 @@
 /* tree.c -- helper functions to build and evaluate the expression tree.
-   Copyright (C) 1990, 91, 92, 93, 94, 2000, 2003, 2004, 2005 Free Software Foundation, Inc.
+   Copyright (C) 1990, 91, 92, 93, 94, 2000, 2003, 2004, 2005, 2006, 2007 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -17,12 +17,14 @@
    USA.
 */
 
-#include "defs.h"
-#include "../gnulib/lib/xalloc.h"
-
 #include <assert.h>
 #include <stdlib.h>
 
+#include "defs.h"
+#include "../gnulib/lib/xalloc.h"
+#include "error.h"
+
+
 #if ENABLE_NLS
 # include <libintl.h>
 # define _(Text) gettext (Text)
@@ -238,10 +240,10 @@ scan_rest (struct predicate **input,
 static boolean
 predicate_is_cost_free(const struct predicate *p)
 {
-  PRED_FUNC pred_func = p->pred_func;
-  
-  if (pred_func == pred_name || pred_func == pred_path ||
-      pred_func == pred_iname || pred_func == pred_ipath)
+  if (pred_is(p, pred_name) ||
+      pred_is(p, pred_path) ||
+      pred_is(p, pred_iname) ||
+      pred_is(p, pred_ipath))
     {
       /* Traditionally (at least 4.1.7 through 4.2.x) GNU find always
        * optimised these cases.
@@ -250,8 +252,10 @@ predicate_is_cost_free(const struct predicate *p)
     }
   else if (options.optimisation_level > 0)
     {
-      if (pred_func == pred_and || pred_func == pred_negate ||
-	       pred_func == pred_comma || pred_func == pred_or)
+      if (pred_is(p, pred_and) ||
+	  pred_is(p, pred_negate) ||
+	  pred_is(p, pred_comma) ||
+	  pred_is(p, pred_or))
 	return false;
       else
 	return NeedsNothing == p->p_cost;
@@ -309,33 +313,6 @@ predlist_insert(struct predlist *list,
     list->tail = list->head;
 }
 
-#if 0
-static void 
-predlist_dump(FILE *fp, const char *label, const struct predlist *list)
-{
-  const struct predicate *p;
-  fprintf(fp, "%s:\n", label);
-  for (p=list->head; p; p=p->pred_left)
-    {
-      print_optlist(fp, p);
-      fprintf(fp, " ");
-    }
-  fprintf(fp, "\n");
-}
-
-static void 
-predlist_merge_nosort(struct predlist *list,
-		      struct predicate **last)
-{
-  if (list->head)
-    {
-      calculate_derived_rates(list->head);
-      merge_pred(list->head, list->tail, last);
-      predlist_init(list);
-    }
-}
-
-#endif
 static int
 pred_cost_compare(const struct predicate *p1, const struct predicate *p2, boolean wantfailure)
 {
@@ -565,13 +542,13 @@ consider_arm_swap(struct predicate *p)
 	      fprintf(stderr, "Success rates: l=%f, r=%f\n", succ_rate_l, succ_rate_r);
 	    }
 	  
-	  if (pred_or == p->pred_func)
+	  if (pred_is(p, pred_or))
 	    {
 	      want_swap = succ_rate_r < succ_rate_l;
 	      if (!want_swap)
 		reason = "Operation is OR and right success rate >= left";
 	    }
-	  else if (pred_and == p->pred_func)
+	  else if (pred_is(p, pred_and))
 	    {
 	      want_swap = succ_rate_r > succ_rate_l;
 	      if (!want_swap)
@@ -911,73 +888,6 @@ merge_pred (struct predicate *beg_list, struct predicate *end_list, struct predi
    that a stat is made as late as possible.  Return true if the top node 
    in TREE requires a stat, false if not. */
 
-static boolean
-mark_stat (struct predicate *tree)
-{
-  /* The tree is executed in-order, so walk this way (apologies to Aerosmith)
-     to find the first predicate for which the stat is needed. */
-  switch (tree->p_type)
-    {
-    case NO_TYPE:
-    case PRIMARY_TYPE:
-      return tree->need_stat;
-
-    case UNI_OP:
-      if (mark_stat (tree->pred_right))
-	tree->need_stat = true;
-      return (false);
-
-    case BI_OP:
-      /* ANDs and ORs are linked along ->left ending in NULL. */
-      if (tree->pred_left != NULL)
-	mark_stat (tree->pred_left);
-
-      if (mark_stat (tree->pred_right))
-	tree->need_stat = true;
-
-      return (false);
-
-    default:
-      error (1, 0, _("oops -- invalid expression type in mark_stat!"));
-      return (false);
-    }
-}
-
-/* Find the first node in expression tree TREE that we will
-   need to know the file type, if any.   Operates in the same 
-   was as mark_stat().
-*/
-static boolean
-mark_type (struct predicate *tree)
-{
-  /* The tree is executed in-order, so walk this way (apologies to Aerosmith)
-     to find the first predicate for which the type information is needed. */
-  switch (tree->p_type)
-    {
-    case NO_TYPE:
-    case PRIMARY_TYPE:
-      return tree->need_type;
-
-    case UNI_OP:
-      if (mark_type (tree->pred_right))
-	tree->need_type = true;
-      return false;
-
-    case BI_OP:
-      /* ANDs and ORs are linked along ->left ending in NULL. */
-      if (tree->pred_left != NULL)
-	mark_type (tree->pred_left);
-
-      if (mark_type (tree->pred_right))
-	tree->need_type = true;
-
-      return false;
-
-    default:
-      error (1, 0, _("oops -- invalid expression type in mark_type!"));
-      return (false);
-    }
-}
 
 struct pred_cost_lookup
 {
@@ -990,7 +900,7 @@ static struct pred_cost_lookup costlookup[] =
     { pred_and       ,  NeedsNothing,        },
     { pred_anewer    ,  NeedsStatInfo,       },
     { pred_atime     ,  NeedsStatInfo,       },
-    { pred_close     ,  NeedsNothing         },
+    { pred_closeparen,  NeedsNothing         },
     { pred_cmin      ,  NeedsStatInfo,       },
     { pred_cnewer    ,  NeedsStatInfo,       },
     { pred_comma     ,  NeedsNothing,        },
@@ -1024,7 +934,7 @@ static struct pred_cost_lookup costlookup[] =
     { pred_nouser    ,  NeedsStatInfo        }, /* true for amortised cost if caching is on */
     { pred_ok        ,  NeedsUserInteraction },
     { pred_okdir     ,  NeedsUserInteraction },
-    { pred_open	     ,  NeedsNothing         },
+    { pred_openparen ,  NeedsNothing         },
     { pred_or        ,  NeedsNothing,        },
     { pred_path	     ,  NeedsNothing         },
     { pred_perm	     ,  NeedsStatInfo        },
@@ -1086,11 +996,10 @@ cost_table_comparison(const void *p1, const void *p2)
 }
 
 static enum EvaluationCost
-get_pred_cost(struct predicate *p)
+get_pred_cost(const struct predicate *p)
 {
   enum EvaluationCost data_requirement_cost = NeedsNothing;
   enum EvaluationCost inherent_cost = NeedsUnknown;
-  PRED_FUNC f = p->pred_func;
 
   if (p->need_stat)
     {
@@ -1105,14 +1014,14 @@ get_pred_cost(struct predicate *p)
       data_requirement_cost = NeedsNothing;
     }
   
-  if (pred_exec == f || pred_execdir == f)
+  if (pred_is(p, pred_exec) || pred_is(p, pred_execdir))
     {
       if (p->args.exec_vec.multiple)
 	inherent_cost = NeedsEventualExec;
       else
 	inherent_cost = NeedsImmediateExec;
     }
-  else if (pred_fprintf == f)
+  else if (pred_is(p, pred_fprintf))
     {
       /* the parser calculated the cost for us. */
       inherent_cost = p->p_cost;
@@ -1215,7 +1124,7 @@ calculate_derived_rates(struct predicate *p)
 
     case UNI_OP:
       /* Unary operators must have exactly one operand */
-      assert(pred_negate == p->pred_func);
+      assert(pred_is(p, pred_negate));
       assert(NULL == p->pred_left);
       p->est_success_rate = (1.0 - p->pred_right->est_success_rate);
       return p->est_success_rate;
@@ -1224,15 +1133,15 @@ calculate_derived_rates(struct predicate *p)
       {
 	float rate;
 	/* Binary operators must have two operands */
-	if (pred_and == p->pred_func)
+	if (pred_is(p, pred_and))
 	  {
 	    rate = getrate(p->pred_right) * getrate(p->pred_left);
 	  }
-	else if (pred_comma == p->pred_func)
+	else if (pred_is(p, pred_comma))
 	  {
 	    rate = 1.0f;
 	  }
-	else if (pred_or == p->pred_func)
+	else if (pred_is(p, pred_or))
 	  {
 	    rate = getrate(p->pred_right) + getrate(p->pred_left);
 	  }
@@ -1304,7 +1213,7 @@ build_expression_tree(int argc, char *argv[], int end_of_leading_options)
   assert(entry_close != NULL);
   assert(entry_print != NULL);
   
-  parse_open (entry_open, argv, &argc);
+  parse_openparen (entry_open, argv, &argc);
   last_pred->p_name = "(";
   predicates->artificial = true;
   parse_begin_user_args(argv, argc, last_pred, predicates);
@@ -1345,7 +1254,7 @@ build_expression_tree(int argc, char *argv[], int end_of_leading_options)
 		   * predicate.  It must be invalid, or not tasty.
 		   */
 		  error (1, 0, _("invalid predicate `%s'"),
-			 argv[i], predicate_name);
+			 predicate_name);
 		}
 	      else
 		{
@@ -1397,7 +1306,7 @@ build_expression_tree(int argc, char *argv[], int end_of_leading_options)
   else
     {
       /* `( user-supplied-expression ) -print'. */
-      parse_close (entry_close, argv, &argc);
+      parse_closeparen (entry_close, argv, &argc);
       last_pred->p_name = ")";
       last_pred->artificial = true;
       pred_sanity_check(last_pred);
@@ -1428,10 +1337,10 @@ build_expression_tree(int argc, char *argv[], int end_of_leading_options)
   if (cur_pred != NULL)
     {
       /* cur_pred->p_name is often NULL here */
-      if (cur_pred->pred_func == pred_close)
+      if (pred_is(cur_pred, pred_closeparen))
 	{
 	  /* e.g. "find \( -true \) \)" */
-	  error (1, 0, _("you have too many ')'"), cur_pred->p_name);
+	  error (1, 0, _("you have too many ')'"));
 	}
       else
 	{
@@ -1460,11 +1369,6 @@ build_expression_tree(int argc, char *argv[], int end_of_leading_options)
   
   /* Check that the tree is still in normalised order */
   check_normalization(eval_tree, true);
-  
-  /* Determine the point, if any, at which to stat the file. */
-  mark_stat (eval_tree);
-  /* Determine the point, if any, at which to determine file type. */
-  mark_type (eval_tree);
 
   if (options.debug_options & (DebugExpressionTree|DebugTreeOpt))
     {
@@ -1477,6 +1381,16 @@ build_expression_tree(int argc, char *argv[], int end_of_leading_options)
 
   return eval_tree;
 }
+
+/* Initialise the performance data for a predicate. 
+ */
+static void
+init_pred_perf(struct predicate *pred)
+{
+  struct predicate_performance_info *p = &pred->perf;
+  p->visits = p->successes = 0;
+}
+
 
 /* Return a pointer to a new predicate structure, which has been
    linked in as the last one in the predicates list.
@@ -1524,6 +1438,7 @@ get_new_pred (const struct parser_table *entry)
   last_pred->literal_control_chars = options.literal_control_chars;
   last_pred->artificial = false;
   last_pred->est_success_rate = 1.0;
+  init_pred_perf(last_pred);
   return last_pred;
 }
 
