@@ -27,12 +27,18 @@
 #ifdef HAVE_SYS_UTSNAME_H
 #include <sys/utsname.h>
 #endif
-
 #include <sys/time.h>
+#include <ctype.h>
+#include <string.h>
+#include <limits.h>
+#include <assert.h>
+
 #include "xalloc.h"
 #include "quotearg.h"
 #include "timespec.h"
-
+#include "error.h"
+#include "verify.h"
+#include "openat.h"
 
 #if ENABLE_NLS
 # include <libintl.h>
@@ -47,12 +53,6 @@
 # define N_(String) String
 #endif
 
-#include <ctype.h>
-#include <string.h>
-#include <limits.h>
-#include <assert.h>
-#include <openat.h>
-
 
 struct debug_option_assoc
 {
@@ -66,6 +66,7 @@ static struct debug_option_assoc debugassoc[] =
     { "tree", DebugExpressionTree, "Display the expression tree" },
     { "search",DebugSearch, "Navigate the directory tree verbosely" },
     { "stat", DebugStat, "Trace calls to stat(2) and lstat(2)" },
+    { "rates", DebugSuccessRates, "Indicate how often each predicate succeeded" },
     { "opt",  DebugExpressionTree|DebugTreeOpt, "Show diagnostic information relating to optimisation" },
     { "exec", DebugExec,  "Show diagnostic information relating to -exec, -execdir, -ok and -okdir" }
   };
@@ -195,11 +196,21 @@ get_statinfo (const char *pathname, const char *name, struct stat *p)
   if (!state.have_stat)
     {
       set_stat_placeholders(p);
-      if ((*options.xstat) (name, p) != 0)
+      if (0 == (*options.xstat) (name, p))
+	{
+	  if (00000 == p->st_mode)
+	    {
+	      /* Savannah bug #16378. */
+	      error(0, 0, _("Warning: file %s appears to have mode 0000"),
+		    quotearg_n_style(0, options.err_quoting_style, name));
+	    }
+	}
+      else
 	{
 	  if (!options.ignore_readdir_race || (errno != ENOENT) )
 	    {
-	      error (0, errno, "%s", pathname);
+	      error (0, errno, "%s",
+		     safely_quote_err_filename(0, pathname));
 	      state.exit_status = 1;
 	    }
 	  return -1;
@@ -208,6 +219,7 @@ get_statinfo (const char *pathname, const char *name, struct stat *p)
   state.have_stat = true;
   state.have_type = true;
   state.type = p->st_mode;
+
   return 0;
 }
 
@@ -284,7 +296,7 @@ do_complete_pending_execdirs(struct predicate *p, int dirfd)
   
   do_complete_pending_execdirs(p->pred_left, dirfd);
   
-  if (p->pred_func == pred_execdir || p->pred_func == pred_okdir)
+  if (pred_is(p, pred_execdir) || pred_is(p, pred_okdir))
     {
       /* It's an exec-family predicate.  p->args.exec_val is valid. */
       if (p->args.exec_vec.multiple)
@@ -333,7 +345,8 @@ complete_pending_execs(struct predicate *p)
   /* It's an exec-family predicate then p->args.exec_val is valid
    * and we can check it. 
    */
-  if (p->pred_func == pred_exec && p->args.exec_vec.multiple)
+  /* XXX: what about pred_ok() ? */
+  if (pred_is(p, pred_exec) && p->args.exec_vec.multiple)
     {
       struct exec_val *execp = &p->args.exec_vec;
       
@@ -365,6 +378,53 @@ cleanup(void)
     }
 }
 
+/* Savannah bug #16378 manifests as an assertion failure in pred_type()
+ * when an NFS server returns st_mode with value 0 (of course the stat(2)
+ * system call is itself returning 0 in this case). 
+ */
+#undef DEBUG_SV_BUG_16378
+#if defined(DEBUG_SV_BUG_16378)
+static int hook_fstatat(int fd, const char *name, struct stat *p, int flags)
+{
+  static int warned = 0;
+
+  if (!warned)
+    {
+      /* No use of _() here; no point asking translators to translate a debug msg */
+      error(0, 0,
+	    "Warning: some debug code is enabled for Savannah bug #16378; "
+	    "this should not occur in released versions of findutils!");
+      warned = 1;
+    }
+  
+  if (0 == strcmp(name, "./mode0file")
+      || 0 == strcmp(name, "mode0file")) 
+    {
+      time_t now = time(NULL);
+      long day = 86400;
+      
+      p->st_rdev = 0;
+      p->st_dev = 0x300;
+      p->st_ino = 0;
+      p->st_mode = 0;		/* SV bug #16378 */
+      p->st_nlink = 1;
+      p->st_uid = geteuid();
+      p->st_gid = 0;
+      p->st_size = 42;
+      p->st_blksize = 32768;
+      p->st_atime = now-1*day;
+      p->st_mtime = now-2*day;
+      p->st_ctime = now-3*day;
+
+      return 0;
+    }
+  return fstatat(fd, name, p, flags);
+}
+
+# undef  fstatat
+# define fstatat(fd,name,p,flags) hook_fstatat((fd),(name),(p),(flags))
+#endif
+
 
 static int
 fallback_stat(const char *name, struct stat *p, int prev_rv)
@@ -407,6 +467,8 @@ fallback_stat(const char *name, struct stat *p, int prev_rv)
 int 
 optionh_stat(const char *name, struct stat *p)
 {
+  if (AT_FDCWD != state.cwd_dir_fd)
+    assert(state.cwd_dir_fd >= 0);
   set_stat_placeholders(p);
   if (0 == state.curdepth) 
     {
@@ -436,6 +498,9 @@ int
 optionl_stat(const char *name, struct stat *p)
 {
   int rv;
+  if (AT_FDCWD != state.cwd_dir_fd)
+    assert(state.cwd_dir_fd >= 0);
+  
   set_stat_placeholders(p);
   rv = fstatat(state.cwd_dir_fd, name, p, 0);
   if (0 == rv)
@@ -451,6 +516,7 @@ optionl_stat(const char *name, struct stat *p)
 int 
 optionp_stat(const char *name, struct stat *p)
 {
+  assert((state.cwd_dir_fd >= 0) || (state.cwd_dir_fd==AT_FDCWD));
   set_stat_placeholders(p);
   return fstatat(state.cwd_dir_fd, name, p, AT_SYMLINK_NOFOLLOW);
 }
@@ -630,7 +696,7 @@ process_debug_options(char *arg)
       if (i >= N_DEBUGASSOC)
 	{
 	  error(0, 0, _("Ignoring unrecognised debug flag %s"),
-		quotearg_n_style(0, locale_quoting_style, arg));
+		quotearg_n_style(0, options.err_quoting_style, arg));
 	}
       p = strtok_r(NULL, delimiters, &token_context);
     }
@@ -828,6 +894,8 @@ set_option_defaults(struct options *p)
 #endif
 
   set_follow_state(SYMLINK_NEVER_DEREF); /* The default is equivalent to -P. */
+
+  p->err_quoting_style = locale_quoting_style;
 }
 
 
@@ -839,3 +907,49 @@ int get_start_dirfd(void)
 {
   return starting_desc;
 }
+
+/* apply_predicate
+ *
+ */
+boolean
+apply_predicate(const char *pathname, struct stat *stat_buf, struct predicate *p)
+{
+  ++p->perf.visits;
+
+  if (p->need_stat || p->need_type)
+    {
+      /* We may need a stat here. */
+      if (get_info(pathname, stat_buf, p) != 0)
+	    return false;
+    }
+  if ((p->pred_func)(pathname, stat_buf, p))
+    {
+      ++(p->perf.successes);
+      return true;
+    }
+  else
+    {
+      return false;
+    }
+}
+
+
+/* safely_quote_err_filename
+ *
+ */
+const char *
+safely_quote_err_filename (int n, char const *arg)
+{
+  return quotearg_n_style (n, options.err_quoting_style, arg);
+}
+
+/* fatal_file_error
+ *
+ */
+void
+fatal_file_error(const char *name)
+{
+  error (1, errno, "%s", safely_quote_err_filename(0, name));
+  /*NOTREACHED*/
+  abort();
+}
