@@ -102,7 +102,8 @@ struct buildcmd_control
 enum BC_INIT_STATUS 
   {
     BC_INIT_OK = 0,
-    BC_INIT_ENV_TOO_BIG
+    BC_INIT_ENV_TOO_BIG,
+    BC_INIT_CANNOT_ACCOMODATE_HEADROOM,
   };
 
 extern size_t bc_size_of_environment (void);
@@ -124,7 +125,8 @@ extern void bc_push_arg (const struct buildcmd_control *ctl,
 extern void  bc_init_state(const struct buildcmd_control *ctl,
 			   struct buildcmd_state *state,
 			   void *usercontext);
-extern enum BC_INIT_STATUS bc_init_controlinfo(struct buildcmd_control *ctl);
+extern enum BC_INIT_STATUS bc_init_controlinfo(struct buildcmd_control *ctl,
+					       size_t arglen_headroom);
 extern size_t bc_get_arg_max(void);
 extern void bc_use_sensible_arg_max(struct buildcmd_control *ctl);
 extern void bc_clear_args(const struct buildcmd_control *ctl,
