@@ -48,12 +48,9 @@
 #include <config.h>
 #include <stdio.h>
 #include <sys/types.h>
-
-#if defined(HAVE_STRING_H) || defined(STDC_HEADERS)
 #include <string.h>
-#else
-#include <strings.h>
-#endif
+#include <errno.h>
+
 
 #ifdef STDC_HEADERS
 #include <stdlib.h>
@@ -77,12 +74,18 @@
 #include "locatedb.h"
 #include <getline.h>
 #include "closeout.h"
+#include "xalloc.h"
 #include "gnulib-version.h"
+#include "progname.h"
+#include "error.h"
+
+#ifndef ATTRIBUTE_NORETURN
+# define ATTRIBUTE_NORETURN __attribute__ ((__noreturn__))
+#endif
 
-char *xmalloc PARAMS((size_t));
 
 /* The name this program was run with.  */
-char *program_name;
+const char *program_name;
 
 /* The 128 most common bigrams in the file list, padded with NULs
    if there are fewer.  */
@@ -133,6 +136,26 @@ or     %s most_common_bigrams < file-list > locate-database\n"),
 }
 
 
+static void inerr (const char *filename) ATTRIBUTE_NORETURN;
+static void outerr(void)                 ATTRIBUTE_NORETURN;
+
+static void
+inerr(const char *filename) 
+{
+  error(1, errno, "%s", filename);
+  /*NOTREACHED*/
+  abort();
+}
+
+static void
+outerr(void) 
+{
+  error(1, errno, _("write error"));
+  /*NOTREACHED*/
+  abort();
+}
+
+
 int
 main (int argc, char **argv)
 {
@@ -145,7 +168,7 @@ main (int argc, char **argv)
   FILE *fp;			/* Most common bigrams file.  */
   int line_len;			/* Length of input line.  */
 
-  program_name = argv[0];
+  set_program_name(argv[0]);
   atexit (close_stdout);
 
   bigram[2] = '\0';
@@ -175,7 +198,7 @@ main (int argc, char **argv)
       perror (argv[1]);
       return 1;
     }
-
+  
   pathsize = oldpathsize = 1026; /* Increased as necessary by getline.  */
   path = xmalloc (pathsize);
   oldpath = xmalloc (oldpathsize);
@@ -186,9 +209,14 @@ main (int argc, char **argv)
 
   /* Copy the list of most common bigrams to the output,
      padding with NULs if there are <128 of them.  */
-  fgets (bigrams, 257, fp);
-  fwrite (bigrams, 1, 256, stdout);
-  fclose (fp);
+  if (NULL == fgets (bigrams, 257, fp))
+    inerr(argv[1]);
+  
+  if (256 != fwrite (bigrams, 1, 256, stdout))
+     outerr();
+
+  if (EOF == fclose (fp))
+     inerr(argv[1]);
 
   while ((line_len = getline (&path, &pathsize, stdin)) > 0)
     {
@@ -210,11 +238,17 @@ main (int argc, char **argv)
 	 otherwise, two bytes plus a marker noting that fact.  */
       if (diffcount < -LOCATEDB_OLD_OFFSET || diffcount > LOCATEDB_OLD_OFFSET)
 	{
-	  putc (LOCATEDB_OLD_ESCAPE, stdout);
-	  putw (diffcount + LOCATEDB_OLD_OFFSET, stdout);
+	  if (EOF ==- putc (LOCATEDB_OLD_ESCAPE, stdout))
+	    outerr();
+	  
+	  if (EOF == putw (diffcount + LOCATEDB_OLD_OFFSET, stdout))
+	    outerr();
 	}
       else
-	putc (diffcount + LOCATEDB_OLD_OFFSET, stdout);
+	{
+	  if (EOF == putc (diffcount + LOCATEDB_OLD_OFFSET, stdout))
+	    outerr();
+	}
 
       /* Look for bigrams in the remainder of the path.  */
       for (pp = path + count; *pp != '\0'; pp += 2)
