@@ -185,8 +185,6 @@ static bool stdout_is_a_tty;
 static bool print_quoted_filename;
 static bool results_were_filtered;
 
-/* static char*  slocate_db_pathname = "/var/lib/slocate/slocate.db"; */
-
 static const char *selected_secure_db = NULL;
 
 
@@ -219,15 +217,15 @@ set_max_db_age(const char *s)
       (0 == val && EINVAL == errno))
     {
       error(1, errno,
-	    _("Invalid argument `%s' for option --max-database-age"),
-	    s);
+	    _("Invalid argument %s for option --max-database-age"),
+	    quotearg_n_style(0, locale_quoting_style, s));
     }
   else if (*end)
     {
       /* errno wasn't set, don't print its message */
       error(1, 0,
-	    _("Invalid argument `%s' for option --max-database-age"),
-	    s);
+	    _("Invalid argument %s for option --max-database-age"),
+	    quotearg_n_style(0, locale_quoting_style, s));
     }
   else
     {
@@ -331,15 +329,6 @@ struct locate_stats
 static struct locate_stats statistics;
 
 
-struct stringbuf
-{
-  char *buffer;
-  size_t buffersize;
-  size_t *preqlen;
-};
-static struct stringbuf casebuf;
-
-
 struct regular_expression
 {
   struct re_pattern_buffer regex; /* for --regex */
@@ -584,7 +573,8 @@ visit_locate02_format(struct process_data *procdata, void *context)
        * reading in data which is outside our control, we
        * cannot prevent it.
        */
-      error(1, 0, _("locate database `%s' is corrupt or invalid"), procdata->dbfile);
+      error(1, 0, _("locate database %s is corrupt or invalid"),
+	    quotearg_n_style(0, locale_quoting_style, procdata->dbfile));
     }
  
   /* Overlay the old path with the remainder of the new.  */
@@ -941,10 +931,11 @@ looking_at_slocate_locatedb (const char *filename,
 		   * We don't know how to handle those.
 		   */
 		  error(0, 0,
-			_("locate database `%s' looks like an slocate "
+			_("locate database %s looks like an slocate "
 			  "database but it seems to have security level %c, "
 			  "which GNU findutils does not currently support"),
-			filename, data[1]);
+			quotearg_n_style(0, locale_quoting_style, filename),
+			data[1]);
 		  return 1;
 		}
 	      else
@@ -1033,10 +1024,9 @@ search_one_database (int argc,
 				  &slocate_seclevel))
     {
       error(0, 0,
-	    _("`%s' is an slocate database.  "
-	      "Support for these is new, expect problems for now "
-	      "(you are, after all, using the CVS code)."),
-	    procdata.dbfile);
+	    _("%s is an slocate database.  "
+	      "Support for these is new, expect problems for now."),
+	    quotearg_n_style(0, locale_quoting_style, procdata.dbfile));
       
       /* slocate also uses frcode, but with a different header. 
        * We handle the header here and then work with the data 
@@ -1047,6 +1037,10 @@ search_one_database (int argc,
 	  /* We don't know what those security levels mean, 
 	   * so do nothing further 
 	   */
+	  error(0, 0,
+		_("%s is an slocate database of unsupported security level %d; skipping it."),
+		quotearg_n_style(0, locale_quoting_style, procdata.dbfile),
+		slocate_seclevel);
 	  return 0;
 	}
       else if (slocate_seclevel > 0)
@@ -1072,9 +1066,9 @@ search_one_database (int argc,
 	      if (enable_print || stats)
 		{
 		  error(0, 0,
-			_("`%s' is an slocate database.  "
+			_("%s is an slocate database.  "
 			  "Turning on the '-e' option."),
-			procdata.dbfile);
+			quotearg_n_style(0, locale_quoting_style, procdata.dbfile));
 		}
 	      do_check_existence = ACCEPT_EXISTING;
 	    }
@@ -1105,7 +1099,14 @@ search_one_database (int argc,
 	    {
 	      int more_read = fread (procdata.original_filename + nread, 1,
 				     256 - nread, procdata.fp);
-	      /* XXX: check more_read+nread! */
+	      if ( (more_read + nread) != 256 )
+		{
+		  error(1, 0,
+			_("Old-format locate database %s is "
+			  "too short to be valid"),
+			quotearg_n_style(0, locale_quoting_style, dbfile));
+		  
+		}
 	    }
 	  
 	  for (i = 0; i < 128; i++)
@@ -1260,7 +1261,8 @@ search_one_database (int argc,
   
   if (ferror (procdata.fp))
     {
-      error (0, errno, "%s", procdata.dbfile);
+      error (0, errno, "%s",
+	     quotearg_n_style(0, locale_quoting_style, procdata.dbfile));
       return 0;
     }
   return plimit->items_accepted;
@@ -1396,7 +1398,8 @@ drop_privs(void)
   return 0;
   
  fail:
-  error(1, errno, "%s", what);
+  error(1, errno, "%s",
+	quotearg_n_style(0, locale_quoting_style, what));
   abort();
   kill(0, SIGKILL);
   _exit(1);
@@ -1593,7 +1596,7 @@ dolocate (int argc, char **argv, int secure_db_fd)
       if (secure_db_fd >= 0)
 	{
 	  close(secure_db_fd);
-	  secure_db_fd = 0;
+	  secure_db_fd = -1;
 	}
     }
 
@@ -1670,7 +1673,8 @@ dolocate (int argc, char **argv, int secure_db_fd)
 	      fd = opendb(e);
 	      if (fd < 0)
 		{
-		  error (0, errno, "%s", e);
+		  error (0, errno, "%s",
+			 quotearg_n_style(0, locale_quoting_style, e));
 		  return 0;
 		}
 	    }
@@ -1693,7 +1697,8 @@ dolocate (int argc, char **argv, int secure_db_fd)
       /* Check the database to see if it is old. */
       if (fstat(fd, &st))
 	{
-	  error (0, errno, "%s", e);
+	  error (0, errno, "%s",
+		 quotearg_n_style(0, locale_quoting_style, e));
 	  /* continue anyway */
 	  filesize = (off_t)0;
 	}
@@ -1720,8 +1725,8 @@ dolocate (int argc, char **argv, int secure_db_fd)
 		  /* For example:
 		     warning: database `fred' is more than 8 days old (actual age is 10 days)*/
 		  error (0, 0,
-			 _("warning: database `%s' is more than %d %s old (actual age is %.1f %s)"),
-			 e, 
+			 _("warning: database %s is more than %d %s old (actual age is %.1f %s)"),
+			 quotearg_n_style(0,  locale_quoting_style, e), 
 			 warn_number_units,              _(warn_name_units),
 			 (age/(double)SECONDS_PER_UNIT), _(warn_name_units));
 		}
@@ -1731,7 +1736,8 @@ dolocate (int argc, char **argv, int secure_db_fd)
       fp = fdopen(fd, "r");
       if (NULL == fp)
 	{
-	  error (0, errno, "%s", e);
+	  error (0, errno, "%s",
+		 quotearg_n_style(0, locale_quoting_style, e));
 	  return 0;
 	}
 
@@ -1745,7 +1751,8 @@ dolocate (int argc, char **argv, int secure_db_fd)
       /* Close the databsase (even if it is stdin) */
       if (fclose (fp) == EOF)
 	{
-	  error (0, errno, "%s", e);
+	  error (0, errno, "%s",
+		 quotearg_n_style(0, locale_quoting_style, e));
 	  return 0;
 	}
     }
